from typing import Optional, Sequence

from bs4 import BeautifulSoup

from pymess.config import settings
from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _


def find_banned_tag(html_body: str, banned_tags: Sequence[str] = None) -> Optional[str]:
    banned_tags = banned_tags if banned_tags is not None else settings.EMAIL_TEMPLATE_BANNED_TAGS
    soup = BeautifulSoup(html_body, 'html.parser')
    return next((tag for tag in banned_tags if soup.find(tag)), None)


def raise_error_if_contains_banned_tags(html_body) -> str:
    """
    Raises a `ValidationError` if a `html_body` contains banned tags or attributes. Otherwise returns the `html_body`.
    """
    tag = find_banned_tag(html_body)
    if tag:
        raise ValidationError(
            _('HTML body contains one of banned tag: {}').format(tag))
    return html_body



from django.utils import timezone

from pymess.backend.emails import EmailBackend


class DummyEmailBackend(EmailBackend):
    """
    Dummy e-mail backend used for testing environments. Backend only logs messages to the database.
    """

    def publish_message(self, message):
        self._update_message_after_sending(message, state=message.State.DEBUG, sent_at=timezone.now())

    def pull_message_info(self, message):
        self._update_message(
            message,
            extra_sender_data={**message.extra_sender_data, 'info': {'debug': True}},
            info_changed_at=timezone.now(),
            update_only_changed_fields=True,
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.26 on 2020-01-08 23:02
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0012_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='emailmessage',
            name='external_id',
            field=models.CharField(blank=True, db_index=True, max_length=250, null=True, verbose_name='external ID'),
        ),
    ]



# Generated by Django 3.1 on 2021-03-11 01:03
from django.db import migrations, models

import pymess.models.emails


def assert_email_contents_are_migrated(apps, schema_editor):
    EmailMessage = apps.get_model('pymess', 'EmailMessage')
    assert not EmailMessage.objects.filter(old_content__isnull=False).exists(), (
        'Please update first to version 0.5.0 and run "migrate_email_contents_to_files" command.'
    )


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0023_migration'),
    ]

    operations = [
        migrations.RunPython(assert_email_contents_are_migrated),
        migrations.AlterField(
            model_name='emailmessage',
            name='content_file',
            field=models.FileField(upload_to=pymess.models.emails.generate_content_filename,
                                   verbose_name='content file'),
        ),
        migrations.RemoveField(
            model_name='emailmessage',
            name='old_content',
        ),
    ]



# Generated by Django 2.2.12 on 2020-05-31 19:02

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('pymess', '0018_migration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='dialermessage',
            name='state',
            field=models.IntegerField(
                choices=[(-1, 'waiting'), (0, 'not assigned'), (1, 'ready'), (2, 'rescheduled by dialer'),
                         (3, 'call in progress'), (4, 'hangup'), (5, 'done'), (6, 'rescheduled'),
                         (7, 'listened up complete message'), (8, 'listened up partial message'), (9, 'unreachable'),
                         (10, 'declined'), (11, 'unanswered'), (12, 'unanswered - hangup by dialer'),
                         (13, 'answered - hangup by customer'), (66, 'error message update'), (77, 'debug'),
                         (88, 'error message was not sent')], db_index=True, editable=False, verbose_name='state'),
        ),
        migrations.AlterField(
            model_name='emailmessage',
            name='state',
            field=models.IntegerField(
                choices=[(1, 'waiting'), (2, 'sending'), (3, 'sent'), (4, 'error message was not sent'), (5, 'debug')],
                db_index=True, editable=False, verbose_name='state'),
        ),
        migrations.AlterField(
            model_name='outputsmsmessage',
            name='state',
            field=models.IntegerField(
                choices=[(1, 'waiting'), (2, 'unknown'), (3, 'sending'), (4, 'sent'), (5, 'error message update'),
                         (6, 'debug'), (7, 'delivered'), (8, 'error message was not sent')], db_index=True,
                editable=False, verbose_name='state'),
        ),
        migrations.AlterField(
            model_name='pushnotificationmessage',
            name='state',
            field=models.PositiveIntegerField(
                choices=[(1, 'waiting'), (2, 'sent'), (3, 'error message was not sent'), (4, 'debug')], db_index=True,
                editable=False, verbose_name='state'),
        ),
    ]



import codecs
import os
from typing import Any

from chamber.shortcuts import change_and_save

from pymess.config import settings

__all__ = [
    "change_and_save",
    "get_email_template_body_from_file",
    "SyncEmailTemplates",
]


def get_email_template_body_from_file(email_template_slug, locale):
    base_dir = settings.EMAIL_HTML_DATA_DIRECTORY
    if locale:
        base_dir = os.path.join(settings.EMAIL_HTML_DATA_DIRECTORY, locale)
    with codecs.open(os.path.join(
            base_dir,
            '{}.html'.format(email_template_slug)), 'r', encoding='utf-8-sig') as file:
        return file.read()


class SyncEmailTemplates:

    def __init__(self, template_slugs=None, locale=None):
        self.template_slugs = template_slugs
        self.locale = locale

    def __call__(self, apps, schema_editor):
        email_template_class = apps.get_model(*settings.EMAIL_TEMPLATE_MODEL.split('.'))
        email_template_qs = email_template_class.objects.all()
        if self.template_slugs:
            email_template_qs = email_template_qs.filter(slug__in=self.template_slugs, locale=self.locale)
        email_templates_found = list(email_template_qs)
        if self.template_slugs:
            email_templates_not_found = set(self.template_slugs) - {t.slug for t in email_templates_found}
            assert not email_templates_not_found, f"Email templates {email_templates_not_found} were not found."
        for email_template in email_templates_found:
            email_template.body = get_email_template_body_from_file(email_template.slug, self.locale)
        email_template_class.objects.bulk_update(email_templates_found, ['body'])



import re

from django.utils import timezone as tz
from django.utils.translation import gettext as _

from pymess.backend.dialer import DialerBackend
from pymess.config import settings
from pymess.enums import DialerMessageState
from pymess.utils.logged_requests import generate_session


class DaktelaDialerBackend(DialerBackend):
    """
    Dialer backend implementing Daktela service https://www.daktela.com/api/v6/models/campaignsrecords
    """

    SESSION_SLUG = 'pymess-daktela_autodialer'
    config = {
        'ACCESS_TOKEN':  None,
        'AUTODIALER_RECORD_TYPE': None,
        'PREDICTIVE_RECORD_TYPE': None,
        'URL': None,
        'STATES_MAPPING': {
            '0': 0,
            '1': 1,
            '2': 2,
            '3': 3,
            '4': 4,
            '5': 5,
            '6': 6,
        },
        'TIMEOUT': 5,  # 5s
    }

    def _get_dialer_api_url(self, name=None):
        name = '/{name}'.format(name=name) if name else ''

        return '{base_url}{name}.json?accessToken={access_token}'.format(
            base_url=self.config['URL'], name=name, access_token=self.config['ACCESS_TOKEN'],
        )

    def _update_dialer_states(self, messages):
        """
        Method uses Daktela API to get info about autodialer call status
        :param messages: list of dialer messages to update
        """
        for message in messages:
            name = message.extra_data['name']
            client_url = self._get_dialer_api_url(name)
            response = generate_session(
                slug=self.SESSION_SLUG,
                related_objects=(message,),
                timeout=self.config['TIMEOUT']
            ).get(client_url)
            resp_json = response.json()
            if response.status_code != 200 and resp_json.get('error'):
                self._update_message_state_with_error(message, error_message=resp_json.get('error'))
                continue

            message.extra_data.update({
                'name': name,
                'daktela_action': resp_json['result']['action'],
                'daktela_statuses': resp_json['result']['statuses'],
            })
            resp_message_state = resp_json['result']['statuses'][0]['name'] if len(
                resp_json['result']['statuses']) else resp_json['result']['action']
            state_mapped = self.config['STATES_MAPPING'][resp_message_state]
            message_error = resp_json['error'] if len(resp_json['error']) else None

            is_final_state = resp_json['result']['action'] == '5'

            try:
                self._update_message(
                    message,
                    state=state_mapped,
                    error=message_error,
                    extra_data=message.extra_data,
                    is_final_state=is_final_state,
                )
            except Exception as ex:
                self._update_message_state_with_error(message, error_message=ex)
                # Do not re-raise caught exception. We do not know exact exception to catch so we catch them all
                # and log them into database. Re-raise exception causes transaction rollback (lost of information about
                # exception).

    def _update_message_state_with_error(self, message, error_message):
        is_final_state = message.number_of_status_check_attempts >= settings.DIALER_NUMBER_OF_STATUS_CHECK_ATTEMPTS
        message_kwargs = {
            'state': DialerMessageState.ERROR_UPDATE,
            'error': ', '.join(error_message) if isinstance(error_message, list) else str(error_message),
            'is_final_state': is_final_state,
        }
        if not is_final_state:
            message_kwargs['number_of_status_check_attempts'] = message.number_of_status_check_attempts + 1
        self._update_message(
            message,
            **message_kwargs
        )

    def publish_message(self, message):
        """
        Method uses Daktela API for sending dialer message
        :param message: dialer message
        """
        client_url = self._get_dialer_api_url()
        try:
            payload = {
                'record_type': (
                    self.config['AUTODIALER_RECORD_TYPE'] if message.is_autodialer
                    else self.config['PREDICTIVE_RECORD_TYPE']
                ),
                'number': re.sub(r'^\+', '00', message.recipient),
                'customFields': {'autodialer_text': [message.content]},
            }
            custom_fields = message.extra_data.get('custom_fields')
            if custom_fields:
                payload['customFields'].update(**custom_fields)

            response = generate_session(
                slug=self.SESSION_SLUG,
                related_objects=(message,),
                timeout=self.config['TIMEOUT']
            ).post(
                client_url,
                json=payload,
            )
            resp_json = response.json()

            message.extra_data.update({
                'name': resp_json['result']['name'],
                'daktela_action': resp_json['result']['action'],
                'daktela_statuses': resp_json['result']['statuses'],
            })

            error_message = resp_json.get('error')
            if error_message:
                self._update_message_after_sending_error(
                    message,
                    error=', '.join(error_message) if isinstance(error_message, list) else str(error_message),
                    state=DialerMessageState.ERROR
                )
            else:
                self._update_message_after_sending(
                    message,
                    state=DialerMessageState.READY,
                    sent_at=tz.now(),
                    extra_data=message.extra_data,
                )
        except Exception as ex:
            self._update_message_after_sending_error(
                message,
                error=str(ex)
            )
            # Do not re-raise caught exception. We do not know exact exception to catch so we catch them all
            # and log them into database. Re-raise exception causes transaction rollback (lost of information about
            # exception).



from http import HTTPStatus
from json.decoder import JSONDecodeError

import requests
from django.utils import timezone
from onesignal import DeviceNotification, OneSignalClient
from onesignal.errors import OneSignalAPIError

from pymess.backend.push import PushNotificationBackend
from pymess.config import settings
from pymess.enums import PushNotificationMessageState
from pymess.utils.logged_requests import generate_session


class OneSignalPushNotificationBackend(PushNotificationBackend):

    config = {
        'APP_ID': None,
        'API_KEY': None,
        'LANGUAGE': None,
        'TIMEOUT': 5,  # 5s
    }

    def _is_result_partial_error(self, result):
        return not result.is_error and result.errors

    def _is_invalid_result(self, result):
        return result.is_error or self._is_result_partial_error(result)

    def publish_message(self, message):
        onesignal_client = OneSignalClient(self.config['APP_ID'],
                                           self.config['API_KEY'])
        onesignal_client.session = generate_session(
            slug='pymess - OneSignal',
            related_objects=(message,),
            timeout=self.config['TIMEOUT']
        )

        languages = {'en'}
        if self.config['LANGUAGE'] is not None:
            languages.add(self.config['LANGUAGE'])
        extra_data = message.extra_data or {}
        if message.redirect_url:
            extra_data['redirectUrl'] = message.redirect_url
        notification = DeviceNotification(
            include_external_user_ids=(message.recipient,),
            contents={language: message.content for language in languages},
            headings={language: message.heading for language in languages},
            data=extra_data,
            url=message.url,
            ios_badge_type=DeviceNotification.IOS_BADGE_TYPE_INCREASE,
            ios_badge_count=1,
        )

        try:
            result = onesignal_client.send(notification)

            extra_sender_data = message.extra_sender_data or {}
            extra_sender_data['result'] = result.body

            if self._is_invalid_result(result):
                self._update_message_after_sending_error(
                    message,
                    state=PushNotificationMessageState.ERROR,
                    error=str(result.errors),
                    extra_sender_data=extra_sender_data,
                )
            else:
                self._update_message_after_sending(
                    message,
                    state=PushNotificationMessageState.SENT,
                    sent_at=timezone.now(),
                    extra_sender_data=extra_sender_data,
                )
        except (JSONDecodeError, requests.exceptions.RequestException, OneSignalAPIError) as ex:
            self._update_message_after_sending_error(
                message, error=str(ex)
            )
            # Do not re-raise caught exception. Re-raise exception causes transaction rollback (loss of information
            # about exception).



# Generated by Django 2.2.15 on 2020-12-17 14:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0020_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='dialermessage',
            name='priority',
            field=models.PositiveSmallIntegerField(default=3, verbose_name='priority'),
        ),
        migrations.AddField(
            model_name='emailmessage',
            name='priority',
            field=models.PositiveSmallIntegerField(default=3, verbose_name='priority'),
        ),
        migrations.AddField(
            model_name='outputsmsmessage',
            name='priority',
            field=models.PositiveSmallIntegerField(default=3, verbose_name='priority'),
        ),
        migrations.AddField(
            model_name='pushnotificationmessage',
            name='priority',
            field=models.PositiveSmallIntegerField(default=3, verbose_name='priority'),
        ),
    ]



from urllib.parse import urlencode

import functools

from django import template
from django.conf import settings
from django.template.base import Node, TemplateSyntaxError, render_value_in_context
from django.template.loader import render_to_string
from django.utils.encoding import force_str
from django.utils.html import format_html

register = template.Library()


class RaiseIfNoneNode(Node):

    def __init__(self, variable_name, email_slug, check_variable, assign_to):
        self.variable_name = variable_name
        self.email_slug = email_slug
        self.check_variable = check_variable
        self.assign_to = assign_to

    def render(self, context):
        value = self.variable_name.resolve(context)
        if not value and not self.check_variable.resolve(context):
            raise ValueError('Email with slug {} missing variable named {}'.format(self.email_slug.resolve(context),
                                                                                   force_str(self.variable_name)))
        if not value:
            value = '{{{{ {} }}}}'.format(force_str(self.variable_name))

        if self.assign_to:
            context[self.assign_to] = render_value_in_context(value, context)
            return ''
        else:
            return render_value_in_context(value, context)

    def __repr__(self):
        return '<SnippetNode>'


@register.tag
def get_or_raise(parser, token):
    bits = token.split_contents()
    assign_to = None
    if len(bits) == 4 and bits[2] == 'as':
        assign_to = bits[3]
    elif len(bits) != 2:
        raise TemplateSyntaxError("%r tag takes one argument: the rendered value" % bits[0])
    variable_name = parser.compile_filter(bits[1])
    check_variable = parser.compile_filter('EMAIL_DISABLE_VARIABLE_VALIDATOR')
    email_slug = parser.compile_filter('EMAIL_SLUG')
    return RaiseIfNoneNode(variable_name, email_slug, check_variable, assign_to)


def check_tag_arguments(f):
    """
    Decorator that checks that all arguments of a simple tag are defined.

    Raises:
        ValueError if any argument is None or empty string.
    """
    @functools.wraps(f)
    def wrapper_decorator(context, **kwargs):
        for k, v in kwargs.items():
            if not context.get('EMAIL_DISABLE_VARIABLE_VALIDATOR') and (v is None or v == ''):
                raise ValueError(
                    'Tag "{tag}" in email slug "{slug}" received an empty argument "{arg}" ({kwargs})'.format(
                        tag=f.__name__,
                        slug=context.get('EMAIL_SLUG'),
                        arg=k,
                        kwargs=kwargs
                    )
                )
        return f(context, **kwargs)
    return wrapper_decorator



class DefaultTimeoutSessionMixin:

    def __init__(self, timeout=None, **kwargs):
        super().__init__(**kwargs)
        self.timeout = timeout

    def request(self, method, url, **kwargs):
        timeout = kwargs.pop('timeout', self.timeout)
        return super().request(method, url, timeout=timeout, **kwargs)


try:
    from security.requests import SecuritySession

    class DefaultTimeoutSecuritySession(DefaultTimeoutSessionMixin, SecuritySession):
        pass

except ImportError:
    from requests import Session

    class DefaultTimeoutSecuritySession(DefaultTimeoutSessionMixin, Session):

        def __init__(self, timeout, slug, related_objects):
            super().__init__(timeout)


def generate_session(slug=None, related_objects=None, timeout=None):
    return DefaultTimeoutSecuritySession(timeout=timeout, slug=slug, related_objects=related_objects)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-03-22 16:12
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('pymess', '0005_push_notifications'),
    ]

    operations = [
        migrations.AddField(
            model_name='dialertemplate',
            name='is_active',
            field=models.BooleanField(default=True, verbose_name='is active'),
        ),
        migrations.AddField(
            model_name='emailtemplate',
            name='is_active',
            field=models.BooleanField(default=True, verbose_name='is active'),
        ),
        migrations.AddField(
            model_name='smstemplate',
            name='is_active',
            field=models.BooleanField(default=True, verbose_name='is active'),
        ),
        migrations.AlterField(
            model_name='dialermessage',
            name='is_final_state',
            field=models.BooleanField(default=False, verbose_name='is final state'),
        ),
        migrations.AlterField(
            model_name='pushnotificationtemplate',
            name='is_active',
            field=models.BooleanField(default=True, verbose_name='is active'),
        ),
    ]



# Generated by Django 3.1.4 on 2020-12-21 18:49

import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0021_migration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='dialermessage',
            name='extra_data',
            field=models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                   null=True, verbose_name='extra data'),
        ),
        migrations.AlterField(
            model_name='dialermessage',
            name='extra_sender_data',
            field=models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                   null=True, verbose_name='extra sender data'),
        ),
        migrations.AlterField(
            model_name='emailmessage',
            name='extra_data',
            field=models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                   null=True, verbose_name='extra data'),
        ),
        migrations.AlterField(
            model_name='emailmessage',
            name='extra_sender_data',
            field=models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                   null=True, verbose_name='extra sender data'),
        ),
        migrations.AlterField(
            model_name='outputsmsmessage',
            name='extra_data',
            field=models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                   null=True, verbose_name='extra data'),
        ),
        migrations.AlterField(
            model_name='outputsmsmessage',
            name='extra_sender_data',
            field=models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                   null=True, verbose_name='extra sender data'),
        ),
        migrations.AlterField(
            model_name='pushnotificationmessage',
            name='extra_data',
            field=models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                   null=True, verbose_name='extra data'),
        ),
        migrations.AlterField(
            model_name='pushnotificationmessage',
            name='extra_sender_data',
            field=models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                   null=True, verbose_name='extra sender data'),
        ),
    ]



from chamber.exceptions import PersistenceException

from pymess.backend import BaseBackend, send_template as _send_template, BaseController
from pymess.config import (
    ControllerType, get_email_template_model, is_turned_on_email_batch_sending, settings,
)
from pymess.models import EmailMessage


class EmailController(BaseController):
    """Controller class for E-mail delegating message to correct E-mail backend"""

    model = EmailMessage
    backend_type_name = ControllerType.EMAIL

    class EmailSendingError(Exception):
        pass

    def get_batch_size(self):
        return settings.EMAIL_BATCH_SIZE

    def get_batch_max_seconds_to_send(self):
        return settings.EMAIL_BATCH_MAX_SECONDS_TO_SEND

    def get_initial_email_state(self, recipient):
        """
        returns initial state for logged e-mail message.
        :param recipient: e-mail address of the recipient
        """
        return self.model.State.WAITING

    def create_message(self, sender, sender_name, recipient, subject, pre_header, content, related_objects, tag, template,
                       attachments, priority=settings.DEFAULT_MESSAGE_PRIORITY, **kwargs):
        """
        Create e-mail which will be logged in the database.
        :param sender: e-mail address of the sender
        :param sender_name: friendly name of the sender
        :param recipient: e-mail address of the receiver
        :param subject: subject of the e-mail message
        :param pre_header: pre header of the e-mail message
        :param content: content of the e-mail message
        :param related_objects: list of related objects that will be linked with the e-mail message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content, subject and sender of the message was created
        :param attachments: list of files that will be sent with the message as attachments
        :param priority: priority of sending message 1 (highest) to 3 (lowest)
        :param kwargs: extra data that will be saved in JSON format in the extra_data model field
        """
        try:
            message = super().create_message(
                recipient=recipient,
                content=content,
                related_objects=related_objects,
                tag=tag,
                template=template,
                sender=sender,
                sender_name=sender_name,
                subject=subject,
                pre_header=pre_header,
                state=self.get_initial_email_state(recipient),
                priority=priority,
                extra_data=kwargs,
                **self.get_backend(recipient).get_extra_message_kwargs()
            )
            if attachments:
                message.attachments.create_from_tripples(*attachments)
            return message
        except PersistenceException as ex:
            raise self.EmailSendingError(str(ex))

    def is_turned_on_batch_sending(self):
        return is_turned_on_email_batch_sending()


class EmailBackend(BaseBackend):
    """
    Base class for E-mail backend containing implementation of e-mail service that
    is used for sending messages.
    """

    def get_batch_max_number_of_send_attempts(self):
        return settings.EMAIL_BATCH_MAX_NUMBER_OF_SEND_ATTEMPTS

    def get_retry_sending(self):
        return settings.EMAIL_RETRY_SENDING and is_turned_on_email_batch_sending()

    def pull_message_info(self, message):
        """
        Pull message info from email service and store into extra_sender_data
        :param message: Email message
        """
        raise NotImplementedError


def send_template(recipient, slug, context_data, locale=None, related_objects=None, attachments=None,
                  tag=None, send_immediately=False):
    """
    Helper for building and sending e-mail message from a template.
    :param recipient: e-mail address of the receiver
    :param slug: slug of the e-mail template
    :param context_data: dict of data that will be sent to the template renderer
    :param related_objects: list of related objects that will be linked with the e-mail message with generic
        relation
    :param attachments: list of files that will be sent with the message as attachments
    :param tag: string mark that will be saved with the message
    :param send_immediately: publishes the message regardless of the `is_turned_on_batch_sending` result
    :return: e-mail message object or None if template cannot be sent
    """
    return _send_template(
        recipient=recipient,
        slug=slug,
        context_data=context_data,
        locale=locale,
        related_objects=related_objects,
        tag=tag,
        template_model=get_email_template_model(),
        attachments=attachments,
        send_immediately=send_immediately
    )


def send(sender, recipient, subject, content, pre_header=None, sender_name=None, related_objects=None, attachments=None, tag=None,
         send_immediately=False, message_backend=None, **kwargs):
    """
    Helper for sending e-mail message.
    :param sender: e-mail address of the sender
    :param recipient: e-mail address of the receiver
    :param subject: subject of the e-mail message
    :param content: content of the e-mail message
    :param pre_header: pre header of the e-mail message
    :param sender_name: friendly name of the sender
    :param related_objects: list of related objects that will be linked with the e-mail message with generic
        relation
    :param tag: string mark that will be saved with the message
    :param attachments: list of files that will be sent with the message as attachments
    :param send_immediately: publishes the message regardless of the `is_turned_on_batch_sending` result
    :param message_backend: message backend instance (if not specified controller will choose the backend)
    :param kwargs: extra data that will be saved in JSON format in the extra_data model field
    :return: True if e-mail was successfully sent or False if e-mail is in error state
    """
    return EmailController().send(
        sender=sender,
        recipient=recipient,
        subject=subject,
        content=content,
        pre_header=pre_header,
        sender_name=sender_name,
        related_objects=related_objects,
        tag=tag,
        attachments=attachments,
        send_immediately=send_immediately,
        message_backend=message_backend,
        **kwargs
    ).failed



from collections import OrderedDict
import re

from enum import Enum, auto

from django.apps import apps
from django.conf import settings as django_settings
from django.utils.module_loading import import_string


DEFAULT_SENDER_BACKEND_NAME = 'default'


class ControllerType(Enum):

    SMS = auto()
    EMAIL = auto()
    DIALER = auto()
    PUSH_NOTIFICATION = auto()


DEFAULTS = {
    # SMS configuration
    'SMS_BACKENDS': {
        DEFAULT_SENDER_BACKEND_NAME: {
            'backend': 'pymess.backend.sms.dummy.DummySMSBackend'
        }
    },
    'SMS_DEFAULT_SENDER_BACKEND_NAME': DEFAULT_SENDER_BACKEND_NAME,
    'SMS_BACKEND_ROUTER': 'pymess.backend.routers.DefaultBackendRouter',
    'SMS_TEMPLATE_MODEL': 'pymess.SMSTemplate',
    'SMS_USE_ACCENT': False,
    'SMS_DEFAULT_PHONE_CODE': None,
    'SMS_LOG_IDLE_MESSAGES': True,
    'SMS_SET_ERROR_TO_IDLE_MESSAGES': True,
    'SMS_IDLE_MESSAGES_TIMEOUT_MINUTES': 10,
    'SMS_BATCH_SENDING': False,
    'SMS_BATCH_SIZE': 20,
    'SMS_BATCH_MAX_NUMBER_OF_SEND_ATTEMPTS': 3,
    'SMS_BATCH_MAX_SECONDS_TO_SEND': 60 * 60,
    'SMS_RETRY_SENDING': True,

    # E-mail configuration
    'EMAIL_BACKENDS': {
        DEFAULT_SENDER_BACKEND_NAME: {
            'backend': 'pymess.backend.emails.dummy.DummyEmailBackend'
        }
    },
    'EMAIL_DEFAULT_SENDER_BACKEND_NAME': DEFAULT_SENDER_BACKEND_NAME,
    'EMAIL_BACKEND_ROUTER': 'pymess.backend.routers.DefaultBackendRouter',
    'EMAIL_TEMPLATE_MODEL': 'pymess.EmailTemplate',
    'EMAIL_TEMPLATE_BASE_TEMPLATE': None,
    'EMAIL_TEMPLATE_TEMPLATETAGS': ['pymess'],
    'EMAIL_TEMPLATE_CONTENT_BLOCK': 'email_content',
    'EMAIL_TEMPLATE_CONTEXT_PROCESSORS': None,
    'EMAIL_TEMPLATE_EXTEND_BODY': True,
    'EMAIL_TEMPLATE_BANNED_TAGS': (
        'applet',
        'amp-iframe',
        'canvas',
        'embed',
        'noscript',
        'object',
        'script',
        'video',
    ),
    'EMAIL_BATCH_SENDING': False,
    'EMAIL_BATCH_SIZE': 20,
    'EMAIL_BATCH_MAX_NUMBER_OF_SEND_ATTEMPTS': 3,
    'EMAIL_BATCH_MAX_SECONDS_TO_SEND': 60 * 60,
    'EMAIL_SENDERS': (),
    'EMAIL_HTML_DATA_DIRECTORY': None,
    'EMAIL_PULL_INFO_BATCH_SIZE': 100,
    'EMAIL_PULL_INFO_DELAY_SECONDS': 60 * 60,  # 1 hour
    'EMAIL_PULL_INFO_MAX_TIMEOUT_FROM_SENT_SECONDS': 60 * 60 * 24 * 30,  # 30 days
    'EMAIL_RETRY_SENDING': True,
    'EMAIL_STORAGE_PATH': 'pymess/emails',

    # Dialer configuration
    'DIALER_BACKENDS': {
        DEFAULT_SENDER_BACKEND_NAME: {
            'backend': 'pymess.backend.dialer.dummy.DummyDialerBackend'
        }
    },
    'DIALER_DEFAULT_SENDER_BACKEND_NAME': DEFAULT_SENDER_BACKEND_NAME,
    'DIALER_BACKEND_ROUTER': 'pymess.backend.routers.DefaultBackendRouter',
    'DIALER_TEMPLATE_MODEL': 'pymess.DialerTemplate',
    'DIALER_IDLE_MESSAGES_TIMEOUT_MINUTES': 60 * 24,
    'DIALER_NUMBER_OF_STATUS_CHECK_ATTEMPTS': 5,
    'DIALER_BATCH_SENDING': False,
    'DIALER_BATCH_SIZE': 20,
    'DIALER_BATCH_MAX_NUMBER_OF_SEND_ATTEMPTS': 3,
    'DIALER_BATCH_MAX_SECONDS_TO_SEND': 60 * 60,
    'DIALER_RETRY_SENDING': True,

    # Push notification settings
    'PUSH_NOTIFICATION_BACKENDS': {
        DEFAULT_SENDER_BACKEND_NAME: {
            'backend': 'pymess.backend.push.dummy.DummyPushNotificationBackend'
        }
    },
    'PUSH_NOTIFICATION_DEFAULT_SENDER_BACKEND_NAME': DEFAULT_SENDER_BACKEND_NAME,
    'PUSH_NOTIFICATION_BACKEND_ROUTER': 'pymess.backend.routers.DefaultBackendRouter',
    'PUSH_NOTIFICATION_TEMPLATE_MODEL': 'pymess.PushNotificationTemplate',
    'PUSH_NOTIFICATION_BATCH_SENDING': False,
    'PUSH_NOTIFICATION_BATCH_SIZE': 20,
    'PUSH_NOTIFICATION_BATCH_MAX_NUMBER_OF_SEND_ATTEMPTS': 3,
    'PUSH_NOTIFICATION_BATCH_MAX_SECONDS_TO_SEND': 60 * 60,
    'PUSH_NOTIFICATION_RETRY_SENDING': True,

    # General settings
    'DEFAULT_MESSAGE_PRIORITY': 3,
    'SHUTDOWN_DETECTION_FUNCTION': None,
}


class BackendNotFound(Exception):
    pass


class Settings(object):
    """
    Pymess settings is loaded lazy like Django settings.
    The reason is usability of override_settings decorator with tests
    """

    def __getattr__(self, attr):
        if attr not in DEFAULTS:
            raise AttributeError('Invalid Pymess setting: "{}"'.format(attr))

        default_value = DEFAULTS[attr]
        value = getattr(django_settings, 'PYMESS_{}'.format(attr), DEFAULTS[attr])

        if isinstance(default_value, dict) and isinstance(value, dict):
            value = {
                **default_value.copy(),
                **value
            }

        return value


settings = Settings()


def get_model(model_name):
    """
    Helper that returns django model class defined by string {app_label}.{model_name}
    """
    return apps.get_model(*model_name.split('.'))


def get_sms_template_model():
    """
    Function returns SMS template model defined in Pymess settings
    """
    return get_model(settings.SMS_TEMPLATE_MODEL)


def get_email_template_model():
    """
    Function returns e-mail template model defined in Pymess settings
    """
    return get_model(settings.EMAIL_TEMPLATE_MODEL)

def get_router(backend_type):
    router_option_name = '{}_BACKEND_ROUTER'.format(backend_type.name)
    return import_string(getattr(settings, router_option_name))()


def _get_backend_config_dict(backend_type):
    backends_option_name = '{}_BACKENDS'.format(backend_type.name)
    return getattr(settings, backends_option_name)


def get_backend(backend_type, backend_name):
    backend_from_config = _get_backend_config_dict(backend_type)[backend_name]
    return import_string(backend_from_config['backend'])(config=backend_from_config.get('config', {}))


def get_default_sender_backend_name(backend_type):
    backend_default_name = '{}_DEFAULT_SENDER_BACKEND_NAME'.format(backend_type.name)
    return getattr(settings, backend_default_name)


def get_supported_backend_paths(backend_type):
    return [backend_config['backend'] for backend_config_name, backend_config in
            _get_backend_config_dict(backend_type).items()]


def get_dialer_template_model():
    """
    Function returns dialer template model defined in Pymess settings
    """
    return get_model(settings.DIALER_TEMPLATE_MODEL)


def get_push_notification_template_model():
    """
    Function returns push notification template model defined in Pymess settings
    """
    return get_model(settings.PUSH_NOTIFICATION_TEMPLATE_MODEL)


def is_turned_on_email_batch_sending():
    return settings.EMAIL_BATCH_SENDING


def is_turned_on_sms_batch_sending():
    return settings.SMS_BATCH_SENDING


def is_turned_on_push_notification_batch_sending():
    return settings.PUSH_NOTIFICATION_BATCH_SENDING


def is_turned_on_dialer_batch_sending():
    return settings.DIALER_BATCH_SENDING



# Generated by Django 4.2.13 on 2024-07-03 18:01

import django.db.models.deletion
from dateutil.relativedelta import relativedelta
from django.db import migrations, models
from django.utils.timezone import localtime


def create_template_ids_from_slugs(apps, schema_editor):
    DialerMessage = apps.get_model("pymess", "DialerMessage")
    DialerTemplateDisallowedObject = apps.get_model("pymess", "DialerTemplateDisallowedObject")
    DialerTemplate = apps.get_model("pymess", "DialerTemplate")

    EmailTemplate = apps.get_model("pymess", "EmailTemplate")
    EmailMessage = apps.get_model("pymess", "EmailMessage")
    EmailTemplateAttachment = apps.get_model("pymess", "EmailTemplateAttachment")
    EmailTemplateDisallowedObject = apps.get_model("pymess", "EmailTemplateDisallowedObject")

    SMSTemplate = apps.get_model("pymess", "SMSTemplate")
    OutputSMSMessage = apps.get_model("pymess", "OutputSMSMessage")

    PushNotificationTemplate = apps.get_model("pymess", "PushNotificationTemplate")
    PushNotificationMessage = apps.get_model("pymess", "PushNotificationMessage")
    SMSTemplateDisallowedObject = apps.get_model("pymess", "SMSTemplateDisallowedObject")

    def migrate_back_template(template, relation, cutoff=False):
        for obj in template.objects.all():
            query = relation.objects.filter(template_old=obj.slug)
            if cutoff:
                query = query.exclude(created_at__lt=localtime() - relativedelta(years=1))
            query.update(template=obj.id)

    migrate_back_template(DialerTemplate, DialerMessage, cutoff=True)
    migrate_back_template(DialerTemplate, DialerTemplateDisallowedObject)

    migrate_back_template(EmailTemplate, EmailMessage, cutoff=True)
    migrate_back_template(EmailTemplate, EmailTemplateAttachment)
    migrate_back_template(EmailTemplate, EmailTemplateDisallowedObject)

    migrate_back_template(SMSTemplate, OutputSMSMessage, cutoff=True)

    migrate_back_template(PushNotificationTemplate, PushNotificationMessage, cutoff=True)
    migrate_back_template(PushNotificationTemplate, SMSTemplateDisallowedObject)


class Migration(migrations.Migration):

    dependencies = [
        ("pymess", "0029_migration"),
    ]

    operations = [
        # 1. Drop the foreign key constraint.
        migrations.AlterField(
            model_name="dialermessage",
            name="template",
            field=models.SlugField(blank=True, max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name="dialertemplatedisallowedobject",
            name="template",
            field=models.SlugField(blank=True, max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name="emailmessage",
            name="template",
            field=models.SlugField(blank=True, max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name="emailtemplateattachment",
            name="template",
            field=models.SlugField(blank=True, max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name="emailtemplatedisallowedobject",
            name="template",
            field=models.SlugField(blank=True, max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name="outputsmsmessage",
            name="template",
            field=models.SlugField(blank=True, max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name="pushnotificationmessage",
            name="template",
            field=models.SlugField(blank=True, max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name="smstemplatedisallowedobject",
            name="template",
            field=models.SlugField(blank=True, max_length=100, null=True),
        ),
        # 2. Rename 'template' -> 'template_old'
        migrations.RenameField(
            model_name="dialermessage",
            old_name="template",
            new_name="template_old",
        ),
        migrations.RenameField(
            model_name="dialertemplatedisallowedobject",
            old_name="template",
            new_name="template_old",
        ),
        migrations.RenameField(
            model_name="emailmessage",
            old_name="template",
            new_name="template_old",
        ),
        migrations.RenameField(
            model_name="emailtemplateattachment",
            old_name="template",
            new_name="template_old",
        ),
        migrations.RenameField(
            model_name="emailtemplatedisallowedobject",
            old_name="template",
            new_name="template_old",
        ),
        migrations.RenameField(
            model_name="outputsmsmessage",
            old_name="template",
            new_name="template_old",
        ),
        migrations.RenameField(
            model_name="pushnotificationmessage",
            old_name="template",
            new_name="template_old",
        ),
        migrations.RenameField(
            model_name="smstemplatedisallowedobject",
            old_name="template",
            new_name="template_old",
        ),
        # 3. Slug is no longer a primary key.
        migrations.AlterField(
            model_name="dialertemplate",
            name="slug",
            field=models.SlugField(editable=False, max_length=100),
        ),
        migrations.AlterField(
            model_name="emailtemplate",
            name="slug",
            field=models.SlugField(editable=False, max_length=100),
        ),
        migrations.AlterField(
            model_name="pushnotificationtemplate",
            name="slug",
            field=models.SlugField(editable=False, max_length=100),
        ),
        migrations.AlterField(
            model_name="smstemplate",
            name="slug",
            field=models.SlugField(editable=False, max_length=100),
        ),
        # 4. We create a new primary key -> AutoField
        migrations.AddField(
            model_name="dialertemplate",
            name="id",
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="emailtemplate",
            name="id",
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="pushnotificationtemplate",
            name="id",
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="smstemplate",
            name="id",
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False),
            preserve_default=False,
        ),
        # 5. We create a new locale field
        migrations.AddField(
            model_name="dialertemplate",
            name="locale",
            field=models.CharField(blank=True, editable=False, max_length=10, null=True),
        ),
        migrations.AddField(
            model_name="emailtemplate",
            name="locale",
            field=models.CharField(blank=True, editable=False, max_length=10, null=True),
        ),
        migrations.AddField(
            model_name="pushnotificationtemplate",
            name="locale",
            field=models.CharField(blank=True, editable=False, max_length=10, null=True),
        ),
        migrations.AddField(
            model_name="smstemplate",
            name="locale",
            field=models.CharField(blank=True, editable=False, max_length=10, null=True),
        ),
        # 5. slug + locale are unique.
        migrations.AlterUniqueTogether(
            name="dialertemplate",
            unique_together={("slug", "locale")},
        ),
        migrations.AlterUniqueTogether(
            name="emailtemplate",
            unique_together={("slug", "locale")},
        ),
        migrations.AlterUniqueTogether(
            name="pushnotificationtemplate",
            unique_together={("slug", "locale")},
        ),
        migrations.AlterUniqueTogether(
            name="smstemplate",
            unique_together={("slug", "locale")},
        ),
        # 6. We need to create removed `template` fields and return the foreign key constraint.
        migrations.AddField(
            model_name="dialermessage",
            name="template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="dialer_messages",
                to="pymess.dialertemplate",
            ),
        ),
        migrations.AddField(
            model_name="dialertemplatedisallowedobject",
            name="template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="disallowed_objects",
                to="pymess.dialertemplate",
            ),
        ),
        migrations.AddField(
            model_name="emailmessage",
            name="template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="email_messages",
                to="pymess.emailtemplate",
            ),
        ),
        migrations.AddField(
            model_name="emailtemplateattachment",
            name="template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="template_attachments",
                to="pymess.emailtemplate",
            ),
        ),
        migrations.AddField(
            model_name="emailtemplatedisallowedobject",
            name="template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="disallowed_objects",
                to="pymess.emailtemplate",
            ),
        ),
        migrations.AddField(
            model_name="outputsmsmessage",
            name="template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="output_sms_messages",
                to="pymess.smstemplate",
            ),
        ),
        migrations.AddField(
            model_name="pushnotificationmessage",
            name="template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="push_notifications",
                to="pymess.pushnotificationtemplate",
            ),
        ),
        migrations.AddField(
            model_name="smstemplatedisallowedobject",
            name="template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="disallowed_objects",
                to="pymess.smstemplate",
            ),
        ),
        # 7. We need to update `template` field with newly created primary keys.
        migrations.RunPython(create_template_ids_from_slugs),
        # 8. We can safely remove old `template_old` field.
        migrations.RemoveField(
            model_name="dialertemplatedisallowedobject",
            name="template_old",
        ),
        migrations.RemoveField(
            model_name="emailtemplateattachment",
            name="template_old",
        ),
        migrations.RemoveField(
            model_name="emailtemplatedisallowedobject",
            name="template_old",
        ),
        migrations.RemoveField(
            model_name="smstemplatedisallowedobject",
            name="template_old",
        ),
        migrations.RemoveField(
            model_name="dialermessage",
            name="template_old",
        ),
        migrations.RemoveField(
            model_name="emailmessage",
            name="template_old",
        ),
        migrations.RemoveField(
            model_name="outputsmsmessage",
            name="template_old",
        ),
        migrations.RemoveField(
            model_name="pushnotificationmessage",
            name="template_old",
        ),
        # 9. We made the field template temporarily nullable so that we can migrate values.
        migrations.AlterField(
            model_name="emailtemplateattachment",
            name="template",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="template_attachments",
                to="pymess.emailtemplate",
            ),
        ),
    ]



import logging
from datetime import timedelta

from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from chamber.exceptions import PersistenceException

from pymess.backend import BaseBackend, BaseController
from pymess.backend import send as _send
from pymess.backend import send_template as _send_template
from pymess.config import (
    ControllerType, get_sms_template_model, get_supported_backend_paths, is_turned_on_sms_batch_sending, settings,
)
from pymess.models import OutputSMSMessage

LOGGER = logging.getLogger(__name__)


class SMSController(BaseController):
    """
    Controller class for SMS delegating message to correct SMS backend
    """
    model = OutputSMSMessage
    backend_type_name = ControllerType.SMS

    class SMSSendingError(Exception):
        pass

    def get_batch_size(self):
        return settings.SMS_BATCH_SIZE

    def get_batch_max_seconds_to_send(self):
        return settings.SMS_BATCH_MAX_SECONDS_TO_SEND

    def get_initial_sms_state(self, recipient):
        """
        returns initial state for logged SMS instance.
        :param recipient: phone number of the recipient
        """
        return self.model.State.WAITING

    def create_message(self, recipient, content, related_objects, tag, template,
                       priority=settings.DEFAULT_MESSAGE_PRIORITY, **kwargs):
        """
        Create SMS which will be logged in the database.
        :param recipient: phone number of the recipient
        :param content: content of the SMS message
        :param related_objects: list of related objects that will be linked with the SMS message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content of the message was created
        :param priority: priority of sending message 1 (highest) to 3 (lowest)
        :param kwargs: extra attributes that will be saved with the message
        """
        try:
            return super().create_message(
                recipient=recipient,
                content=content,
                related_objects=related_objects,
                tag=tag,
                template=template,
                state=self.get_initial_sms_state(recipient),
                priority=priority,
                extra_data=kwargs,
                **self.get_backend(recipient).get_extra_message_kwargs()
            )
        except PersistenceException as ex:
            raise self.SMSSendingError(str(ex))

    def bulk_check_sms_states(self):
        """
        Method that find messages that is not in the final state and updates its states.
        """
        for backend in get_supported_backend_paths(self.backend_type_name):
            messages_to_check = self.model.objects.filter(state=self.model.State.SENDING, backend=backend)
            if messages_to_check.exists():
                import_string(backend)().update_sms_states(messages_to_check)

            idle_output_sms = messages_to_check.filter(
                created_at__lt=timezone.now() - timedelta(minutes=settings.SMS_IDLE_MESSAGES_TIMEOUT_MINUTES),
            )
            if settings.SMS_LOG_IDLE_MESSAGES and idle_output_sms.exists():
                LOGGER.warning('{count_sms} Output SMS is more than {timeout} minutes in state "SENDING"'.format(
                    count_sms=idle_output_sms.count(), timeout=settings.SMS_IDLE_MESSAGES_TIMEOUT_MINUTES
                ))

            if settings.SMS_SET_ERROR_TO_IDLE_MESSAGES:
                idle_output_sms.update(
                    state=self.model.State.ERROR, error=_('timeouted')
                )

    def is_turned_on_batch_sending(self):
        return is_turned_on_sms_batch_sending()


class SMSBackend(BaseBackend):
    """
    Base class for SMS backend containing implementation of concrete SMS service that
    is used for sending SMS messages.
    """

    def update_sms_states(self, messages):
        """
        If SMS sender provides check SMS delivery this method can be overridden.
        :param messages: messages which state will be updated
        """
        raise NotImplementedError('Check SMS state is not supported with the backend')

    def get_batch_max_number_of_send_attempts(self):
        return settings.SMS_BATCH_MAX_NUMBER_OF_SEND_ATTEMPTS

    def get_retry_sending(self):
        return settings.SMS_RETRY_SENDING and is_turned_on_sms_batch_sending()


def send_template(recipient, slug, context_data, locale=None, related_objects=None, tag=None, send_immediately=False):
    """
    Helper for building and sending SMS message from a template.
    :param recipient: phone number of the recipient
    :param slug: slug of a SMS template
    :param context_data: dict of data that will be sent to the template renderer
    :param related_objects: list of related objects that will be linked with the SMS message using generic
        relation
    :param tag: string mark that will be saved with the message
    :return: SMS message object or None if template cannot be sent
    :param send_immediately: publishes the message regardless of the `is_turned_on_batch_sending` result
    """
    return _send_template(
        recipient=recipient,
        slug=slug,
        locale=locale,
        context_data=context_data,
        related_objects=related_objects,
        tag=tag,
        template_model=get_sms_template_model(),
        send_immediately=send_immediately
    )


def send(recipient, content, related_objects=None, tag=None, send_immediately=False, **kwargs):
    """
    Helper for sending SMS message.
    :param recipient: phone number of the recipient
    :param content: text content of the messages
    :param related_objects:
    :param tag: string mark that will be saved with the message
    :param kwargs: extra attributes that will be stored with messages
    :param send_immediately: publishes the message regardless of the `is_turned_on_batch_sending` result
    :return: True if SMS was successfully sent or False if message is in error state
    """
    return _send(
        recipient=recipient,
        content=content,
        related_objects=related_objects,
        tag=tag,
        message_controller=SMSController(),
        send_immediately=send_immediately,
        **kwargs
    )



import import_string
import logging

from chamber.utils.transaction import smart_atomic
from django.core.management.base import BaseCommand, CommandError
from django.db import DatabaseError

from pymess.backend.dialer import DialerController
from pymess.backend.emails import EmailController
from pymess.backend.push import PushNotificationController
from pymess.backend.sms import SMSController
from pymess.config import settings

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Command for sending messages in batch.
    """

    controllers = {
        'email': EmailController(),
        'push-notification': PushNotificationController(),
        'dialer': DialerController(),
        'sms': SMSController()
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.touched_message_pks = set()
        self.send_message_pks = set()
        self.failed_message_pks = set()
        self.shutdown_detection_function = (
            import_string(settings.SHUTDOWN_DETECTION_FUNCTION) if settings.SHUTDOWN_DETECTION_FUNCTION else None
        )

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument('--type', action='store', dest='type', default='email',
                            help='Tells Django what type of messages should be send '
                                 '(email/push-notification/dialer/sms).')

    @smart_atomic
    def _send_message(self, controller):
        message = controller.get_waiting_or_retry_messages().exclude(
            pk__in=self.touched_message_pks
        ).select_for_update(
            nowait=True
        ).order_by('priority', 'created_at').first()
        if not message or not controller.is_turned_on_batch_sending():
            return False

        self.touched_message_pks.add(message.pk)
        try:
            if controller.publish_or_retry_message(message):
                self.send_message_pks.add(message.pk)
            else:
                self.failed_message_pks.add(message.pk)
        except Exception as ex:
            # Rollback should not be applied for already send e-mails
            logger.exception(ex)
            self.failed_message_pks.add(message.pk)
        return True

    def _print_result(self, title, message_pks):
        if message_pks:
            self.stdout.write('{}: {} ({})'.format(title, len(message_pks), ', '.join((str(pk) for pk in message_pks))))
        else:
            self.stdout.write('{}: {}'.format(title, len(message_pks)))

    def _is_system_shutting_down(self):
        return self.shutdown_detection_function is not None and self.shutdown_detection_function()

    def handle(self, type, *args, **options):
        controller = self.controllers[type]
        if not controller.is_turned_on_batch_sending():
            raise CommandError('Batch sending is turned off')

        try:
            for _ in range(controller.get_batch_size()):
                if self._is_system_shutting_down():
                    self.stdout.write("System is shutting down, exitting gracefully.")
                    break
                if not self._send_message(controller):
                    break
            self._print_result('sent messages', self.send_message_pks)
            self._print_result('failed messages', self.failed_message_pks)
        except DatabaseError:
            self.stdout.write('messages are already locked')
